import builtins
import sys
from textwrap import dedent

import pytest

from reprexlite.code import CodeBlock
from reprexlite.formatting import venues_dispatcher
from reprexlite.reprex import reprex
from tests.expected_reprexes import (
    ASSETS_DIR,
    INPUT,
    MOCK_VERSION,
    MockDateTime,
    MockSessionInfo,
    expected_reprexes,
)


@pytest.fixture
def patch_datetime(monkeypatch):
    monkeypatch.setattr(sys.modules["reprexlite.formatting"], "datetime", MockDateTime)


@pytest.fixture
def patch_version(monkeypatch):
    monkeypatch.setattr(sys.modules["reprexlite.formatting"], "__version__", MOCK_VERSION)


@pytest.fixture
def patch_session_info(monkeypatch):
    monkeypatch.setattr(sys.modules["reprexlite.formatting"], "SessionInfo", MockSessionInfo)


@pytest.fixture
def no_pygments(monkeypatch):
    import_orig = builtins.__import__

    def mocked_import(name, *args):
        if name.startswith("pygments"):
            raise ImportError()
        return import_orig(name, *args)

    monkeypatch.setattr(builtins, "__import__", mocked_import)


@pytest.mark.parametrize("ereprex", expected_reprexes, ids=[e.filename for e in expected_reprexes])
def test_reprex(ereprex, patch_datetime, patch_session_info, patch_version):
    actual = reprex(INPUT, **ereprex.kwargs, print_=False)
    with (ASSETS_DIR / ereprex.filename).open("r") as fp:
        assert str(actual) + "\n" == fp.read()


def test_html(patch_datetime, patch_version, no_pygments):
    reprex_class = venues_dispatcher["html"]
    code_block = CodeBlock(INPUT)
    reprex = reprex_class(code_block)
    print(reprex)
    expected = dedent(
        """
        <pre><code>x = 2
        x + 2
        #> 4</code></pre>
        <p><sup>Created at DATETIME by <a href="https://github.com/jayqi/reprexlite">reprexlite</a> vVERSION</sup></p>
        """
    ).strip()
    assert str(reprex) == expected


def test_old_results():
    input = dedent(
        """\
        arr = [1, 2, 3, 4, 5]
        [x + 1 for x in arr]
        #> old line
        """
    )

    assert "#> old line" not in str(reprex(input))  # old_results False (default)
    assert "#> old line" in str(reprex(input, old_results=True))  # old_results True



import platform

from reprexlite.session_info import SessionInfo


def test_session_info():
    session_info = str(SessionInfo())
    assert session_info
    assert platform.python_version() in session_info
    assert "pytest" in session_info

    lines = session_info.split("\n")
    for line in lines:
        assert len(line) <= 80



from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
import shutil
import sys
from textwrap import dedent
from typing import Any, Dict

from tqdm import tqdm

from reprexlite import reprex
from reprexlite.session_info import Package, SessionInfo

ASSETS_DIR = (Path(__file__).parent / "assets").resolve()


INPUT = dedent(
    """\
    x = 2
    x + 2
    """
)


@dataclass
class ExpectedReprex:
    filename: str
    kwargs: Dict[str, Any]


expected_reprexes = [
    # Defaults
    ExpectedReprex("gh.md", {"venue": "gh"}),
    ExpectedReprex("so.md", {"venue": "so"}),
    ExpectedReprex("ds.md", {"venue": "ds"}),
    ExpectedReprex("html.html", {"venue": "html"}),
    ExpectedReprex("py.py", {"venue": "py"}),
    ExpectedReprex("rtf.rtf", {"venue": "rtf"}),
    ExpectedReprex("slack.txt", {"venue": "slack"}),
    # With ad
    ExpectedReprex("ad/gh.md", {"venue": "gh", "advertise": True}),
    ExpectedReprex("ad/so.md", {"venue": "so", "advertise": True}),
    ExpectedReprex("ad/ds.md", {"venue": "ds", "advertise": True}),
    ExpectedReprex("ad/html.html", {"venue": "html", "advertise": True}),
    ExpectedReprex("ad/py.py", {"venue": "py", "advertise": True}),
    ExpectedReprex("ad/rtf.rtf", {"venue": "rtf", "advertise": True}),
    ExpectedReprex("ad/slack.txt", {"venue": "slack", "advertise": True}),
    # No ad
    ExpectedReprex("no_ad/gh.md", {"venue": "gh", "advertise": False}),
    ExpectedReprex("no_ad/so.md", {"venue": "so", "advertise": False}),
    ExpectedReprex("no_ad/ds.md", {"venue": "ds", "advertise": False}),
    ExpectedReprex("no_ad/html.html", {"venue": "html", "advertise": False}),
    ExpectedReprex("no_ad/py.py", {"venue": "py", "advertise": False}),
    ExpectedReprex("no_ad/rtf.rtf", {"venue": "rtf", "advertise": False}),
    ExpectedReprex("no_ad/slack.txt", {"venue": "slack", "advertise": False}),
    # With session info
    ExpectedReprex("session_info/gh.md", {"venue": "gh", "session_info": True}),
    ExpectedReprex("session_info/so.md", {"venue": "so", "session_info": True}),
    ExpectedReprex("session_info/ds.md", {"venue": "ds", "session_info": True}),
    ExpectedReprex("session_info/html.html", {"venue": "html", "session_info": True}),
    ExpectedReprex("session_info/py.py", {"venue": "py", "session_info": True}),
    ExpectedReprex("session_info/rtf.rtf", {"venue": "rtf", "session_info": True}),
    ExpectedReprex("session_info/slack.txt", {"venue": "slack", "session_info": True}),
]

MOCK_VERSION = "VERSION"


@contextmanager
def patch_version():
    version = sys.modules["reprexlite.formatting"].__version__
    sys.modules["reprexlite.formatting"].__version__ = MOCK_VERSION
    yield
    sys.modules["reprexlite.formatting"].__version__ = version


class MockDateTime:
    @classmethod
    def now(cls):
        return cls()

    def astimezone(self):
        return self

    def strftime(self, format):
        return "DATETIME"


@contextmanager
def patch_datetime():

    datetime = sys.modules["reprexlite.formatting"].datetime
    sys.modules["reprexlite.formatting"].datetime = MockDateTime
    yield
    sys.modules["reprexlite.formatting"].datetime = datetime


class MockPackage(Package):
    def __init__(self, name: str, version: str):
        self._name = name
        self._version = version

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version


class MockSessionInfo(SessionInfo):
    def __init__(self, *args, **kwargs):
        self.python_version = "3.x.y"
        self.python_build_date = "Jan 01 2020 03:33:33"
        self.os = "GLaDOS"
        self.packages = [
            MockPackage("datatable", "1.0"),
            MockPackage("ggplot2", "2.0"),
            MockPackage("pkgnet", "3.0"),
        ]


@contextmanager
def patch_session_info():
    sys.modules["reprexlite.formatting"].SessionInfo = MockSessionInfo
    yield
    sys.modules["reprexlite.formatting"].SessionInfo = SessionInfo


if __name__ == "__main__":
    shutil.rmtree(ASSETS_DIR, ignore_errors=True)
    with patch_datetime(), patch_version(), patch_session_info():
        for ereprex in tqdm(expected_reprexes):
            outfile = ASSETS_DIR / ereprex.filename
            outfile.parent.mkdir(exist_ok=True)
            reprex(INPUT, outfile=outfile, **ereprex.kwargs, print_=False)



import platform
import sys
from typing import List, Tuple

if sys.version_info[:2] >= (3, 8):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata


class SessionInfo:
    """Class for pretty-formatting Python session info. Includes details about your Python version,
    your operating system, and the Python packages installed in your current environment.

    Attributes:
        python_version (str): Python version for current session
        python_build_date (str): Date
        os (str): OS information for current session
        packages (List[Package]): List of Python packages installed in current virtual environment.
    """

    def __init__(self):
        self.python_version: str = platform.python_version()
        self.python_build_date: str = platform.python_build()[1]

        self.os: str = platform.platform()
        self.packages: List[Package] = [
            Package(distr) for distr in importlib_metadata.Distribution.discover()
        ]

    def __str__(self):
        lines = ["-- Session Info --" + "-" * 60]
        lines += tabulate(
            [
                ("version", f"Python {self.python_version} ({self.python_build_date})"),
                ("os", self.os),
            ]
        )
        lines += ["-- Packages --" + "-" * 64]
        lines += tabulate([(pkg.name, pkg.version) for pkg in sorted(self.packages)])
        return "\n".join(lines).strip()


class Package:
    """Interface for adapting [`importlib.metadata.Distribution`](https://docs.python.org/3/library/importlib.metadata.html#distributions)
    instances for introspection by [`SessionInfo`][reprexlite.session_info.SessionInfo].
    """

    def __init__(self, distribution: importlib_metadata.Distribution):
        self.distribution = distribution

    @property
    def name(self) -> str:
        return self.distribution.metadata["Name"]

    @property
    def version(self) -> str:
        return self.distribution.version

    def __lt__(self, other) -> bool:
        if not isinstance(other, Package):
            raise ValueError
        return self.name < other.name


def tabulate(rows: List[Tuple[str, str]]) -> List[str]:
    """Utility function for printing a two-column table as text with whitespace padding.

    Args:
        rows (List[Tuple[str, str]]): Rows of table as tuples of (left cell, right cell)

    Returns:
        Rows of table formatted as strings with whitespace padding
    """
    left_max = max(len(row[0]) for row in rows)
    out = []
    for left, right in rows:
        padding = (left_max + 1 - len(left)) * " "
        out.append(left + padding + right)
    return out



from contextlib import redirect_stdout
from io import StringIO
from itertools import chain
from pprint import pformat
import traceback
from typing import Any, List, Optional, Union

import libcst as cst

NO_RETURN = object()
"""Explicit placeholder object for statements, which have no return value (as opposed to
expressions)."""


class Result:
    """Class that holds the result of evaluated code. Use `str(...)` on an instance to produce a
    pretty-formatted comment block representation of the result.

    Attributes:
        result (Any): Some Python object, intended to be the return value of evaluated Python code.
        comment (str): Line prefix to use when rendering the result for a reprex.
    """

    def __init__(self, result: Any, stdout: Optional[str] = None, comment: str = "#>"):
        self.result = result
        self.stdout = stdout
        self.comment = comment

    def __str__(self) -> str:
        lines = []
        if self.stdout:
            lines.extend(self.stdout.split("\n"))
        if self.result is not NO_RETURN and (self.result is not None or not self.stdout):
            # NO_RETURN -> don't print
            # None and stdout -> don't print
            # None and no stdout -> print
            # Anything else -> print
            lines.extend(pformat(self.result, indent=2, width=77).split("\n"))
        return "\n".join(f"{self.comment} " + line for line in lines)

    def __bool__(self) -> bool:
        # If result is NO_RETURN and blank stdout, nothing to print
        return self.result is not NO_RETURN or bool(self.stdout)


class Statement:
    """Class that holds a LibCST parsed statement. The evaluate method will evaluate the statement
    and return a [`Result`][reprexlite.code.Result] object. To recover the original source code
    for an instancement, call `str(...)` on it. You can optionally autoformat the returned source
    code, controlled by the `style` attribute.

    Attributes:
        stmt (Union[libcst.SimpleStatementLine, libcst.BaseCompoundStatement]): LibCST parsed
            statement.
        style (bool): Whether to autoformat the source code with black.
    """

    def __init__(
        self, stmt: Union[cst.SimpleStatementLine, cst.BaseCompoundStatement], style: bool = False
    ):
        self.stmt = stmt
        self.style = style

    def evaluate(self, scope: dict) -> Result:
        stdout_io = StringIO()
        try:
            with redirect_stdout(stdout_io):
                try:
                    result = eval(str(self).strip(), scope, scope)
                except SyntaxError:
                    exec(str(self).strip(), scope, scope)
                    result = NO_RETURN
            stdout = stdout_io.getvalue().strip()
        except Exception:
            result = NO_RETURN
            stdout = traceback.format_exc().strip()
        finally:
            stdout_io.close()
        return Result(result, stdout=stdout)

    def __str__(self) -> str:
        code = cst.Module(body=[self.stmt]).code
        if self.style:
            try:
                from black import Mode, format_str
            except ImportError:
                raise ImportError("Must install black to restyle code.")

            code = format_str(code, mode=Mode())
        if code.endswith("\n"):
            # Strip trailing newline without stripping deliberate ones.
            code = code[:-1]
        return code


class CodeBlock:
    """Class that takes a block of Python code input and evaluates it. Call `str(...)` on an
    instance to get back a string containing the original source with evaluated outputs embedded
    as comments below each statement.

    Attributes:
        input (str): Block of Python code
        style (bool): Whether to use black to autoformat code in returned string representation.
        comment (str): Line prefix to use when rendering the evaluated results.
        terminal (bool): Whether to apply syntax highlighting to the string representation.
            Requires optional dependency Pygments.
        tree (libcst.Module): Parsed LibCST concrete syntax tree of input code.
        statements (List[Statement]): List of individual statements parsed from input code.
        results (List[Result]): List of evaluated results corresponding to each item of statements.
    """

    def __init__(
        self,
        input: str,
        style: bool = False,
        comment: str = "#>",
        terminal=False,
        old_results: bool = False,
    ):
        """Initializer method.

        Args:
            input (str): Block of Python code
            style (bool): Whether to use black to autoformat code in returned string
                representation. Defaults to False.
            comment (str): Line prefix to use when rendering the evaluated results. Defaults to
                "#>".
            terminal (bool): Whether to apply syntax highlighting to the string representation.
                Requires optional dependency Pygments. Defaults to False.
            old_results (bool): Whether to keep old results, i.e., comment lines in input that
                match the `comment` prefix. False means these lines are removed, in effect meaning
                an inputted regex will have its results regenerated. Defaults to False.
        """
        if not old_results and comment in input:
            input = "\n".join(line for line in input.split("\n") if not line.startswith(comment))
        self.input: str = input
        self.terminal = terminal
        # Parse code
        self.tree: cst.Module = cst.parse_module(input)
        self.statements: List[Statement] = [
            Statement(stmt, style=style) for stmt in self.tree.body
        ]
        # Evaluate code
        self.namespace: dict = {}
        self.results: List[Result] = [stmt.evaluate(self.namespace) for stmt in self.statements]
        for res in self.results:
            res.comment = comment

    def __str__(self):
        header = cst.Module(body=[], header=self.tree.header).code.strip()
        code = "\n".join(
            str(line) for line in chain.from_iterable(zip(self.statements, self.results)) if line
        )
        footer = cst.Module(body=[], footer=self.tree.footer).code.strip()
        out = "\n".join([header, code, footer])
        if self.terminal:
            try:
                from pygments import highlight
                from pygments.formatters import Terminal256Formatter
                from pygments.lexers import PythonLexer

                out = highlight(out, PythonLexer(), Terminal256Formatter(style="friendly"))
            except ImportError:
                pass
        return out.strip()

    def _repr_html_(self):
        out = []
        try:
            from pygments import highlight
            from pygments.formatters import HtmlFormatter
            from pygments.lexers import PythonLexer

            formatter = HtmlFormatter(style="friendly", wrapcode=True)
            out.append(f"<style>{formatter.get_style_defs('.highlight')}</style>")
            out.append(highlight(str(self), PythonLexer(), formatter))
        except ImportError:
            out.append(f"<pre><code>{self}</code></pre>")
        return "\n".join(out)



from pathlib import Path
from typing import Optional

from reprexlite.code import CodeBlock
from reprexlite.formatting import Reprex, venues_dispatcher


def reprex(
    input: str,
    outfile: Optional[Path] = None,
    venue="gh",
    advertise: Optional[bool] = None,
    session_info: bool = False,
    style: bool = False,
    comment: str = "#>",
    old_results: bool = False,
    print_=True,
    terminal=False,
) -> Reprex:
    """Render reproducible examples of Python code for sharing. This function will evaluate your
    code and returns an instance of a [`Reprex`][reprexlite.formatting.Reprex] subclass. Calling
    `str(...)` on the `Reprex` object will return your code with the evaluated results embedded
    as comments, plus additional markup appropriate to the sharing venue set by the `venue` keyword
    argument.

    For example, for the `gh` venue for GitHub Flavored Markdown, you'll get a reprex whose string
    representation looks like:

    ````
    ```python
    x = 2
    x + 2
    #> 4
    ```

    <sup>Created at 2021-02-15 16:58:47 PST by [reprexlite](https://github.com/jayqi/reprexlite) v0.1.0</sup>
    ````

    The supported `venue` formats are:

    - `gh` : GitHub Flavored Markdown
    - `so` : StackOverflow, alias for gh
    - `ds` : Discourse, alias for gh
    - `html` : HTML
    - `py` : Python script
    - `rtf` : Rich Text Format
    - `slack` : Slack

    Args:
        input (str): Block of Python code
        outfile (Optional[Path]): Optional file path to write reprex to. Defaults to None.
        venue (str): Determines the output format by the venue you want to share the code. Defaults
            to "gh" for GitHub Flavored Markdown.
        advertise (Optional[bool]): Whether to include a note that links back to the reprexlite
            package. Default `None` will use the default set by choice of `venue`.
        session_info (bool): Whether to include additional details about your Python version,
            operating system, and installed packages. Defaults to False.
        style (bool): Whether to autoformat your code with black. Defaults to False.
        comment (str): Line prefix to use for displaying evaluated results. Defaults to "#>".
        old_results (bool): Whether to keep old results, i.e., comment lines in input that match
            the `comment` prefix. False means these lines are removed, in effect meaning an
            inputted regex will have its results regenerated. Defaults to False.
        print_ (bool): Whether to print your reprex to console. Defaults to True.
        terminal (bool): Whether to use syntax highlighting for 256-color terminal display.
            Requires optional dependency Pygments. Defaults to False.

    Returns:
        Instance of a `Reprex` concrete subclass for `venue`.
    """

    if outfile or venue in ["html", "rtf"]:
        # Don't screw output file or lexing for HTML and RTF with terminal syntax highlighting
        terminal = False
    code_block = CodeBlock(
        input, style=style, comment=comment, old_results=old_results, terminal=terminal
    )

    reprex = venues_dispatcher[venue](
        code_block=code_block, advertise=advertise, session_info=session_info
    )
    if outfile is not None:
        with outfile.open("w") as fp:
            fp.write(str(reprex) + "\n")
    if print_:
        print(reprex)
    return reprex
